
//			CHARENCODING command:


#ifndef CHARENCODING_COMMAND_VER
#define CHARENCODING_COMMAND_VER		1.0.0.0


#include "HtmlEntities.cpp"


const wchar_t CharEncodingHelpText[]	= L"Displays UTF-8 and UTF-16 encoding of characters.\n\
\n\
CHARENCODING /UTF8 /UTF16 value entity \"string\" ...\r\n\
\n\
\t/UTF8\tshow UTF-8 encoding\n\
\t/UTF16\tshow UTF-16 encoding\n\
\n\
Character values may be decimal, or hexadecimal with a leading 0x.\n\
Entities are as in HTML 4; the leading ampersand may be omitted.\n\
Entities are case sensitive.\n";




BOOL IsEmoji( unsigned int Value )
{
	if ( ( Value >= 0x1f600 ) && ( Value <= 0x1f64f ) )
		return TRUE;

	return FALSE;
}



DLLExports int WINAPI charencoding( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t String[MAX_ARG_LEN]		= L"";
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	BOOL ShowUTF8				= FALSE;
	BOOL ShowUTF16				= FALSE;
	BOOL HighChar				= FALSE;
	unsigned int value			= 0;
	unsigned int val2			= 0;
	unsigned int j				= 0;
	unsigned int k				= 0;

	BYTE b1, b2, b3, b4;



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_RETAIN_QUOTES );



	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( CharEncodingHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( CHARENCODING_COMMAND_VER ) );
			else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
				int m = 2;

				if ( ( towupper( arg[i][m] ) == 'T' ) && ( towupper( arg[i][m + 1] ) == 'F' ) )
					m += 2;

				if ( arg[i][m] == '-' )
					m++;

				if ( !_wcsicmp( arg[i] + m, L"8" ) )
					ShowUTF8 = TRUE;
				else if ( !_wcsicmp( arg[i] + m, L"16" ) )
					ShowUTF16 = TRUE;
				else {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else if ( !_wcsicmp( arg[i], L"crlf" ) || !_wcsicmp( arg[i], L"eol" ) || !wcscmp( arg[i], L"/" ) ) {
				if ( j < ( MAX_ARG_LEN - 2 ) ) {
					String[j++] = '\r';
					String[j++] = '\n';
					String[j]	= '\0';
				}
				else
					goto ErrorOverflow;
			}
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( iswdigit( arg[i][0] ) ) {
				if ( ParseInt( arg[i], &value, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ( value < 1 ) || ( value > 0x10ffff ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Value must be 0x0001 - 0x10FFFF: 0x%04X\r\n", value );
					return 1;
				}

				k = ( value <= 0xffff ) ? 1 : 2;

				if ( ( j + k ) >= ( MAX_ARG_LEN - 1 ) )
					goto ErrorOverflow;

				if ( k == 2 ) {
					MM_CreateSurrogatePair( value, String + j );
					j += 2;
					String[j]	= '\0';
				}
				else {
					String[j++]	= value;
					String[j]	= '\0';
				}
			}
			else if ( ( towupper( arg[i][0] ) == 'U' ) && ( arg[i][1] == '+' ) && ( iswxdigit( arg[i][2] ) ) ) {
				wchar_t *sep1 = NULL;

				value = wcstoul( arg[i] + 2, &sep1, 16 );

				if ( *sep1 != '\0' ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad code point: \"%s\"\r\n", arg[i] );
					return -1;
				}

				if ( ( value < 1 ) || ( value > 0x10ffff ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Value must be U+0001 - U+FFFF: U+%04X\r\n", value );
					return 1;
				}
				
				k = ( value <= 0xffff ) ? 1 : 2;

				if ( ( j + k ) >= ( MAX_ARG_LEN - 1 ) )
					goto ErrorOverflow;

				if ( k == 2 ) {
					MM_CreateSurrogatePair( value, String + j );
					j += 2;
					String[j]	= '\0';
				}
				else {
					String[j++]	= value;
					String[j]	= '\0';
				}
			}
			else if ( ( arglen[i] > 1 ) && ( arg[i][0] == QUOTE ) && ( arg[i][arglen[i] - 1] == QUOTE ) ) {

				if ( ( j + arglen[i] - 2 ) >= ( MAX_ARG_LEN - 1 ) )
					goto ErrorOverflow;

				for ( int k = 1; k < ( arglen[i] - 1 ); k++ ) {
					String[j++] = arg[i][k];
					if ( ( arg[i][k + 1] == QUOTE ) && arg[i][k + 2] )
						k++;
				}

				String[j] = '\0';
			}
			else {
				value = 0;
				temp[0]		= '\0';

				if ( arg[i][0] != '&' ) {
					temp[0]		= '&';
					temp[1]		= '\0';
				}

				wcscat_s( temp, SHORT_BUF_LEN, arg[i] );
				value = ParseEntity( temp, NULL );

				if ( !value ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown entity: \"%s\"\r\n", temp );
					return 1;
				}

				k = ( value <= 0xffff ) ? 1 : 2;

				if ( ( j + k ) >= ( MAX_ARG_LEN - 1 ) )
					goto ErrorOverflow;

				if ( k == 2 ) {
					MM_CreateSurrogatePair( value, String + j );
					j += 2;
					String[j]	= '\0';
				}
				else {
					String[j++]	= value;
					String[j]	= '\0';
				}
			}

		}
	}


	if ( !ShowUTF8 && !ShowUTF16 )
		ShowUTF8 = ShowUTF16 = TRUE;



	QPuts( L"Character   " );

	if ( ShowUTF8 )
		QPuts( L"  UTF-8       " );

	if ( ShowUTF16 )
		QPuts( L"  UTF-16" );

	CrLf();

	QPuts( L"------------" );

	if ( ShowUTF8 )
		QPuts( L"  ----------- " );

	if ( ShowUTF16 )
		QPuts( L"  ----------" );

	CrLf();

	for ( wchar_t *ch = String; *ch; ch++ ) {
		HighChar = FALSE;
		value = *ch;

		if ( ( value & 0xfc00 ) == 0xd800 ) {
			if ( ( ch[1] & 0xfc00 ) == 0xdc00 ) {
				value = ( value & 0x03ff ) << 10;
				value = ( value | ( ch[1] & 0x03ff ) ) + 0x10000;
				HighChar = TRUE;
			}
		}


		Printf( L"U+%04X", value );

		if ( !HighChar ) {
			val2 = value;

			if ( value < 0x20 )
				val2 = 0x20;
			else if ( ( value >= 0x80 ) && ( value < 0xa0 ) )
				val2 = 0x20;
			else if ( ( value >= 0xd800 ) && ( value <= 0xdfff ) )
				val2 = 0xb7;

			Printf( L"    %c ", val2 );
		}
		else if ( value < 0x100000 ) {
			if ( IsEmoji( value ) )
				Printf( L"  %c%c ", ch[0], ch[1] );
			else
				QPuts( L"     " );
		}
		else
			QPuts( L"    " );



		if ( ShowUTF8 ) {

			if ( value < 0x0080 )
				Printf( L"  %02X          ", value );
			else if ( value < 0x0800 ) {
				b1	= ( value >> 6 ) | 0xc0;
				b2	= ( value & 0x3f ) | 0x80;
				Printf( L"  %02X %02X       ", b1, b2 );
			}
			else if ( value < 0x10000 ) {
				b1	= ( value >> 12 ) | 0xe0;
				b2	= ( ( value & 0x0fff ) >> 6 ) | 0x80;
				b3	= ( value & 0x3f ) | 0x80;
				Printf( L"  %02X %02X %02X    ", b1, b2, b3 );
			}
			else {
				b1	= ( value >> 18 ) | 0xf0;
				b2	= ( ( value & 0x03ffff ) >> 12 ) | 0x80;
				b3	= ( ( value & 0x0fff ) >> 6 ) | 0x80;
				b4	= ( value & 0x03f ) | 0x80;
				Printf( L"  %02X %02X %02X %02X ", b1, b2, b3, b4 );
			}
		}


		if ( ShowUTF16 ) {
			if ( HighChar )
				Printf( L"  %04X  %04X", ch[0], ch[1] );
			else
				Printf( L"  %04X      ", value );
		}


		if ( HighChar )
			ch++;

		CrLf();
	}


	return 0;


ErrorOverflow:

	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Resulting string is too long\r\n" );
	return 1;
}


#endif	//	CHARENCODING_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-27

*/
