
//			COPYCHARS command:


#ifndef COPYCHARS_COMMAND_VER
#define COPYCHARS_COMMAND_VER		1.0.0.0


#include "HtmlEntities.cpp"


const wchar_t CopyCharsHelpText[]	= L"Put characters on the clipboard.\n\
\n\
COPYCHARS /A /Q value entity \"string\" ...\r\n\
\n\
\t/A\tappend\n\
\t/Q\tquietly\n\
\n\
Character values may be decimal, or hexadecimal with a leading 0x.\n\
Entities are as in HTML 4; the leading ampersand may be omitted.\n\
Entities are case sensitive.\n";




DLLExports int WINAPI copychars( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t String[MAX_ARG_LEN]		= L"";
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	BOOL Append					= FALSE;
	BOOL Quiet					= FALSE;
	BOOL Verbose				= FALSE;

	unsigned int value			= 0;
	unsigned int j				= 0;
	unsigned int k				= 0;
	size_t nChars				= 0;

	size_t BufferSize			= 0;
	HGLOBAL MemHandle			= NULL;
	wchar_t *Buffer				= NULL;

	LPTSTR OldClipboardText		= NULL;
	size_t OldClipboardLen		= 0;
	size_t OldClipboardChars	= 0;

	int rv						= 2;



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_RETAIN_QUOTES );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcsncmp( arg[i], L"/?",2 ) )
				return ShowCmdHelp( CopyCharsHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( COPYCHARS_COMMAND_VER ) );
			else if ( !_wcsicmp( arg[i], L"/A" ) )
				Append = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Quiet = TRUE;
			else if ( !_wcsicmp( arg[i], L"/V" ) )
				Verbose = TRUE;
			else if ( !_wcsicmp( arg[i], L"crlf" ) || !_wcsicmp( arg[i], L"eol" ) || !wcscmp( arg[i], L"/" ) ) {
				if ( j < ( MAX_ARG_LEN - 2 ) ) {
					String[j++] = '\r';
					String[j++] = '\n';
					String[j] = '\0';
				}
				else
					goto ErrorOverflow;
			}
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( iswdigit( arg[i][0] ) ) {
				if ( ParseInt( arg[i], &value, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ( value < 1 ) || ( value > 0x10ffff ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Value must be 0x0001 - 0x10FFFF: 0x%04X\r\n", value );
					return 1;
				}

				k = ( value <= 0xffff ) ? 1 : 2;

				if ( ( j + k ) >= ( MAX_ARG_LEN - 1 ) )
					goto ErrorOverflow;

				if ( k == 2 ) {
					MM_CreateSurrogatePair( value, String + j );
					j += 2;
					String[j]	= '\0';
				}
				else {
					String[j++]	= value;
					String[j]	= '\0';
				}
			}
			else if ( ( towupper( arg[i][0] ) == 'U' ) && ( arg[i][1] == '+' ) && ( iswxdigit( arg[i][2] ) ) ) {
				wchar_t *sep1 = NULL;

				value = wcstoul( arg[i] + 2, &sep1, 16 );

				if ( *sep1 != '\0' ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad code point: \"%s\"\r\n", arg[i] );
					return -1;
				}

				if ( ( value < 1 ) || ( value > 0x10ffff ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Value must be U+0001 - U+FFFF: U+%04X\r\n", value );
					return 1;
				}
				
				k = ( value <= 0xffff ) ? 1 : 2;

				if ( ( j + k ) >= ( MAX_ARG_LEN - 1 ) )
					goto ErrorOverflow;

				if ( k == 2 ) {
					MM_CreateSurrogatePair( value, String + j );
					j += 2;
					String[j]	= '\0';
				}
				else {
					String[j++]	= value;
					String[j]	= '\0';
				}
			}
			else if ( ( arglen[i] > 1 ) && ( arg[i][0] == QUOTE ) && ( arg[i][arglen[i] - 1] == QUOTE ) ) {

				if ( ( j + arglen[i] - 2 ) >= ( MAX_ARG_LEN - 1 ) )
					goto ErrorOverflow;

				for ( int k = 1; k < ( arglen[i] - 1 ); k++ ) {
					String[j++] = arg[i][k];
					if ( ( arg[i][k + 1] == QUOTE ) && arg[i][k + 2] )
						k++;
				}

				String[j] = '\0';
			}
			else {
				value = 0;
				temp[0]		= '\0';

				if ( arg[i][0] != '&' ) {
					temp[0]		= '&';
					temp[1]		= '\0';
				}

				wcscat_s( temp, SHORT_BUF_LEN, arg[i] );
				value = ParseEntity( temp, NULL );

				if ( !value ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown entity: \"%s\"\r\n", temp );
					return 1;
				}

				k = ( value <= 0xffff ) ? 1 : 2;

				if ( ( j + k ) >= ( MAX_ARG_LEN - 1 ) )
					goto ErrorOverflow;

				if ( k == 2 ) {
					MM_CreateSurrogatePair( value, String + j );
					j += 2;
					String[j]	= '\0';
				}
				else {
					String[j++]	= value;
					String[j]	= '\0';
				}
			}
		}
	}


	if ( !j ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No characters to copy\r\n" );
		return 1;
	}


	if ( Append ) {
		OldClipboardText = GetClipboardText( &OldClipboardLen );

		if ( OldClipboardText && !Quiet )
			OldClipboardChars = MM_CodePointsIn( OldClipboardText );
	}


	if ( !OpenClipboard( GetConsoleWindow() ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening clipboard\r\n" );
		free( OldClipboardText );
		return 2;
	}


	BufferSize = sizeof( wchar_t ) * ( wcslen( String ) + OldClipboardLen + 1 );
	rv = 2;

	MemHandle = GlobalAlloc( GHND, BufferSize );
	if ( MemHandle ) {

		Buffer = (wchar_t*) GlobalLock( MemHandle );
		if ( Buffer ) {

			if ( EmptyClipboard() ) {

				if ( OldClipboardText ) {
					memcpy( Buffer, OldClipboardText, OldClipboardLen * sizeof( wchar_t ) );
					free( OldClipboardText );
				}

				memcpy( Buffer + OldClipboardLen, String, ( wcslen( String ) + 1 ) * sizeof( wchar_t ) );

				if ( SetClipboardData( CF_UNICODETEXT, MemHandle ) )
					rv = 0;
				else {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Error setting clipboard data\r\n" );
				}
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error emptying the clipboard\r\n" );
			}
		}
		else {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error locking memory handle\r\n" );
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error allocating memory for buffer\r\n" );
	}

	CloseClipboard();

	if ( !rv ) {
		GlobalFree( MemHandle );

		if ( !Quiet ) {
			nChars = MM_CodePointsIn( String );			//	 + OldClipboardChars;
			Printf( L"Copied %s character%s to the clipboard.\r\n", Commafy( nChars ), ( nChars == 1 ) ? L"" : L"s" );

			if ( Verbose )
				Printf( L"   \"%s\"\r\n", String );
		}
	}

	return rv;


ErrorOverflow :

	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Resulting string is too long\r\n" );
	return 1;
}


#endif	//	COPYCHARS_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-27
				Now uses HtmlEntities.cpp.  /?? now shows this file's
				version info.

*/
