
//			DEHTML command, @MKENTITIES function :


#ifndef DEHTML_COMMAND_VER
#define DEHTML_COMMAND_VER			1.0.2.1

#include "HtmlEntities.cpp"


#define SH_OPTIONS					0x0003
#define SH_NOSCRIPT					0x0004
#define SH_COMMENTS					0x0008
#define SH_HEADER					0x0010
#define SH_CHECK_META_TAGS			0x0020
#define SH_STRIP_BLANK_LINES		0x0040
#define SH_BODY_ONLY				0x0080
#define SH_REMOVE_TITLE				0x0100

#define NUM_SUFFIX_TAGS				32
#define NUM_CHARSET_NAMES			32

#ifndef PLUGIN_BUFFER_MAX
#define PLUGIN_BUFFER_MAX			32768
#endif	//	PLUGIN_BUFFER_MAX



const wchar_t DeHTMLHelpText[] = L"Strip HTML from a file and dump the contents to stdout.\n\
\n\
DEHTML /B /C /CP:n /E /H /M /N /O:n /P /R /S filename....\r\n\
\n\
\t/B\tbody and title only\n\
\t/C\tinclude text in comments\n\
\t/CP:\tinterpret OEM text as code page n\n\
\t/E\tomit empty lines\n\
\t/H\tdisplay filenames\n\
\t/M\tlook in META tags for charset information\n\
\t/N\tinclude text in NOSCRIPT tags\n\
\t/O:\tinclude OPTION text in SELECT blocks:\n\
\t\t    0: don\u2019t include (default)\n\
\t\t    1: only include the first OPTION\n\
\t\t    2: include all OPTION text\n\
\t/P\tpage output\n\
\t/R\tremove title\n\
\t/S\tsearch in subdirectories for matching files\n\
\n\
Input filenames may be specified on the command line, or text may \n\
be piped or redirected into DEHTML.\n";



wchar_t SuffixTag[NUM_SUFFIX_TAGS][12] = { L"BR", L"TITLE", L"H1", L"H2", L"H3", L"H4", L"H5", L"H6", L"/TR", L"/LI", L"/TABLE",
	L"/UL", L"/OL", L"/DIV", L"/P", L"/ADDRESS", L"/BLOCKQUOTE", L"/TITLE", L"/H1", L"/H2", L"/H3", L"/H4", L"/H5", L"/H6",
	L"/PRE", L"/SELECT", L"/TD", L"/TH", L"TABLE", L"UL", L"OL", L"HR" };

int TagSuffix[NUM_SUFFIX_TAGS] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2,
	1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 3, 3, 4, 4, 4, 5 };


wchar_t CharsetName[NUM_CHARSET_NAMES][14] = { L"windows-1252", L"iso-8859-1", L"windows-1250", L"windows-1251", L"windows-1253", \
		L"windows-1254", L"windows-1255", L"windows-1256", L"windows-1257", L"windows-1258", L"ibm437", L"cp437", \
		L"ibm775", L"cp775", L"ibm850", L"cp850", L"ibm852", L"cp852", L"ibm855", L"cp855", L"ibm857", L"cp857", \
		L"ibm00858", L"cp00858", L"ibm862", L"cp862", L"ibm866", L"cp866", L"windows-874", L"macintosh", L"koi8-r", L"koi8-u" };

wchar_t CharsetCodePage[NUM_CHARSET_NAMES] = { 1252, 1252, 1250, 1251, 1253, 1254, 1255, 1256, 1257, 1258, 437, 437, \
		775, 775, 850, 850, 852, 852, 855, 855, 857, 857, 858, 858, 862, 862, 866, 866, 874, 10000, 20866, 21866 };


struct _DeHtmlArgs {
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};





int DH_DumpEntities()
{
	wchar_t ch			= '\0';


	for ( int i = 0; i < NUM_HTML_ENTITIES; i++ ) {
		ch = HtmlEntity[i].Char;
		Printf( L"%-12s  0x%04X   %c\r\n", HtmlEntity[i].Name, ch, ( iswcntrl( ch ) ) ? 0xb7 : ch );
	}

	CrLf();
	Printf( L"%d entities supported.\r\n", NUM_HTML_ENTITIES );
	return 0;
}



int PrintLongEntity( wchar_t ch, LPTSTR OutBuffer, int *OutBufferIndex )
{
	if ( ( *OutBufferIndex + 5 ) >= ( MAX_ARG_LEN - 1 ) ) {
		QPuts( OutBuffer );
		*OutBufferIndex = 0;
		OutBuffer[*OutBufferIndex] = '\0';
	}

	if ( ch == 0x2026 ) {
		wcscat_s( OutBuffer, MAX_ARG_LEN, L"..." );
		*OutBufferIndex += 3;
		return 1;
	}
	else if ( ch == 0x2122 ) {
		wcscat_s( OutBuffer, MAX_ARG_LEN, L"(TM)" );
		*OutBufferIndex += 4;
		return 1;
	}

	return 0;
}



int CheckMetaTag( _MMFile *MMFile, LPTSTR inString )
{
	if ( !MMFile )
		return 0;
	else if ( !inString )
		return 0;

	if ( _wcsnicmp( inString, L"http-equiv=", 11 ) )
		if ( _wcsnicmp( inString, L"charset=", 8 ) )
			return 0;


	wchar_t *ch = inString;

	while ( *ch && _wcsnicmp( ch, L"charset=", 8 ) )
		ch++;

	if ( *ch == '\0' )
		return 0;

	ch += 8;
	if ( *ch == '\0' )
		return 0;

	wchar_t Charset[SHORT_BUF_LEN] = L"";
	int j = 0;

	for ( ; *ch; ch++ ) {
		if ( ( *ch == QUOTE ) && j )
			break;

		if ( ( *ch == ';' ) && j )
			break;

		if ( *ch == QUOTE )
			continue;

		if ( iswspace( *ch ) && ( j == 0 ) )
			continue;

		if ( j < ( SHORT_BUF_LEN - 1 ) ) {
			Charset[j++] = towlower( *ch );
			Charset[j] = 0x00;
		}
	}

	if ( Charset[0] == '\0' )
		return 0;

	// Printf( L"   * Charset = \"%s\"\r\n", Charset );

	if ( !wcscmp( Charset, L"utf-8" ) ) {
		MMFile->FileEncoding	= FILE_UTF8;
		// Printf( L"   * Switching to UTF-8:\r\n" );
		return 1;
	}

	int TryCodePage = -1;
	for ( int i = 0; i < NUM_CHARSET_NAMES; i++ )
		if ( !wcscmp( Charset, CharsetName[i] ) )
			TryCodePage = i;

	if ( TryCodePage < 0 )
		return 0;

	for ( int i = 0; i < NUM_CODE_PAGES; i++ ) {
		if ( CodePageNumber[i] == TryCodePage ) {
			MMFile->CodePage		= TryCodePage;
			MMFile->CodePageIndex	= i;
			// Printf( L"   * Switching to code page %d:\r\n", TryCodePage );
			return 1;
		}
	}
	
	return 0;
}



size_t CountNonblanks( LPCTSTR String )
{
	if ( !String )
		return 0;


	size_t k		= 0;
	wchar_t *ch		= (wchar_t*) String;


	while ( *ch ) {
		if ( !iswspace( *ch ) && !iswcntrl( *ch ) )
			k++;

		ch++;
	}

	return k;
}



int StripHTML( LPCTSTR inFilename, int Flags, void *MoreArgs )
{
	if ( !inFilename )
		return -1;
	else if ( inFilename[0] == '\0' )
		return -1;


	_MMFile MMFile;

	int InTag = 0;
	wchar_t TagName[SHORT_BUF_LEN]			= L"";
	wchar_t TagAttributes[SHORT_BUF_LEN]	= L"";
	wchar_t OutBuffer[MAX_ARG_LEN]			= L"";
	int OutBufferIndex						= 0;
	BOOL Dirty								= FALSE;
	BOOL Nonblanks							= FALSE;

	int ShowSelectBlocks		= ( Flags & SH_OPTIONS );
	BOOL ShowNoScriptBlocks		= ( ( Flags & SH_NOSCRIPT ) > 0 );
	BOOL ShowComments			= ( ( Flags & SH_COMMENTS ) > 0 );
	BOOL CheckMetaTags			= ( ( Flags & SH_CHECK_META_TAGS ) > 0 );
	BOOL IncludeBlankLines		= ( ( Flags & SH_STRIP_BLANK_LINES ) == 0 );
	BOOL BodyOnly				= ( Flags & SH_BODY_ONLY ) > 0;
	

	BOOL InHtml = FALSE, InHead = FALSE, InBody = FALSE, InTitle = FALSE, InScript = FALSE, InApplet = FALSE;
	BOOL InComment = FALSE, InStyle = FALSE, InPreBlock = FALSE;
	int InSelect = 0;
	
	int ch = '\0', lastch = '\0', pretagch = '\0';
	int j = 0;

	_DeHtmlArgs *args	= (_DeHtmlArgs*) MoreArgs;


	size_t BOMSize = 0;


	//  Open file:

	int rv = OpenMMFile( &MMFile, inFilename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC | MM_FLAG_HIGH_CHARS, args->CodePage, args->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", inFilename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;


	CL_HookBreak();

	if ( MMFile.FileEncoding != FILE_OEM )
		CheckMetaTags = FALSE;


	if ( Flags & SH_HEADER )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;


	//  Loop through the contents of the file:

	ch = SPACE;

	while ( MMFile.Index < MMFile.FileSize ) {
		lastch = ch;

		ch = MMGetChar( &MMFile );

		if ( ch < 0 )
			break;						//  abort on error or eof
		else if ( ch == '\0' )
			continue;					//  ignore any nulls in input text

		if ( InComment ) {
			if ( ( ch == '-' ) && ( lastch == '-' ) ) {
				if ( MMPeekAhead( &MMFile, 0 ) == '>' ) {
					InComment = FALSE;
					MMGetChar( &MMFile );
				}

				if ( ShowComments && OutBufferIndex ) {
					OutBufferIndex--;
					OutBuffer[OutBufferIndex] = '\0';
				}

				continue;
			}

			if ( !ShowComments )
				continue;
		}

		if ( ch == '<' ) {
			if ( ( iswspace( MMPeekAhead( &MMFile, 0 ) ) == 0 ) && ( InTag == 0 ) ) {
				InTag = 1;
				j = 0;
				TagName[j] = '\0';
				TagAttributes[j] = '\0';
				pretagch = lastch;
				continue;
			}
		}
		else if ( InTag && ( ch == '>' ) ) {
			InTag = 0;
			ch = pretagch;

			if ( TagName[0] ) {
				if ( !wcscmp( TagName, L"HTML" ) )
					InHtml = TRUE;
				else if ( !wcscmp( TagName, L"/HTML" ) )
					InHtml = InHead = InBody = FALSE;

				if ( InHtml ) {
					if ( !wcscmp( TagName, L"HEAD" ) )
						InHead = TRUE;
					else if ( !wcscmp( TagName, L"/HEAD" ) )
						InHead = FALSE;
					else if ( !wcscmp( TagName, L"BODY" ) )
						InBody = TRUE;
					else if ( !wcscmp( TagName, L"/BODY" ) )
						InBody = FALSE;
				}

				if ( InHead ) {
					if ( !wcscmp( TagName, L"TITLE" ) )
						InTitle = TRUE;
					else if ( !wcscmp( TagName, L"/TITLE" ) ) {
						InTitle = FALSE;

						if ( Flags & SH_REMOVE_TITLE ) {
							OutBufferIndex	= 0;
							OutBuffer[OutBufferIndex]	= '\0';
						}
					}
				}

				if ( !wcscmp( TagName, L"SCRIPT" ) )
					InScript = TRUE;
				else if ( !wcscmp( TagName, L"/SCRIPT" ) )
					InScript = FALSE;
				else if ( !wcscmp( TagName, L"NOSCRIPT" ) && !ShowNoScriptBlocks )
					InScript = TRUE;
				else if ( !wcscmp( TagName, L"/NOSCRIPT" ) )
					InScript = FALSE;
				else if ( !wcscmp( TagName, L"APPLET" ) && !ShowNoScriptBlocks )
					InApplet = TRUE;
				else if ( !wcscmp( TagName, L"/APPLET" ) )
					InApplet = FALSE;
				else if ( !wcscmp( TagName, L"SELECT" ) )
					InSelect = ShowSelectBlocks;
				else if ( !wcscmp( TagName, L"/SELECT" ) )
					InSelect = 0;
				else if ( !wcscmp( TagName, L"/OPTION" ) && ( InSelect == 1 ) )
					InSelect = 2;
				else if ( !wcscmp( TagName, L"STYLE" ) )
					InStyle = TRUE;
				else if ( !wcscmp( TagName, L"/STYLE" ) )
					InStyle = FALSE;
				else if ( !wcscmp( TagName, L"PRE" ) )
					InPreBlock = TRUE;
				else if ( !wcscmp( TagName, L"/PRE" ) )
					InPreBlock = FALSE;

				if ( CheckMetaTags && InHead && !wcscmp( TagName, L"META" ) )
					if ( CheckMetaTag( &MMFile, TagAttributes ) == 1 )
						CheckMetaTags = FALSE;

				int q = 0;
				for ( int i = 0; i < NUM_SUFFIX_TAGS; i++ )
					if ( wcscmp( TagName, SuffixTag[i] ) == 0 )
						q = TagSuffix[i];

				if ( !wcscmp( TagName, L"/OPTION" ) && !ShowSelectBlocks )
					q = 1;

				switch ( q ) {

				case 1 :

					//	open block :
					if ( OutBufferIndex ) {
						if ( CountNonblanks( OutBuffer ) )
							Nonblanks = TRUE;

						if ( Nonblanks || IncludeBlankLines ) {
							QPuts( OutBuffer );
							Dirty = TRUE;
						}

						OutBufferIndex = 0;
						OutBuffer[OutBufferIndex] = '\0';
					}

					if ( Dirty )
						if ( CL_EndLine() == USER_ABORT )
							goto UserAbort;

					ch			= SPACE;
					Dirty		= FALSE;
					Nonblanks	= FALSE;
					break;

				case 2 :

					//	close block :
					if ( OutBufferIndex ) {
						if ( CountNonblanks( OutBuffer ) )
							Nonblanks = TRUE;

						if ( Nonblanks || IncludeBlankLines ) {
							QPuts( OutBuffer );
							Dirty = TRUE;
						}

						OutBufferIndex = 0;
						OutBuffer[OutBufferIndex] = '\0';
					}

					if ( Dirty ) {
						if ( CL_EndLine( IncludeBlankLines ? 2 : 1 ) == USER_ABORT )
							goto UserAbort;
					}

					ch			= SPACE;
					Dirty		= FALSE;
					Nonblanks	= FALSE;
					break;

				case 3 :
					if ( OutBufferIndex >= ( MAX_ARG_LEN - 16 ) ) {
						QPuts( OutBuffer );
						OutBufferIndex = 0;
						OutBuffer[OutBufferIndex] = '\0';
						Dirty = TRUE;
					}

					OutBuffer[OutBufferIndex++]	= TAB;
					OutBuffer[OutBufferIndex]	= '\0';
					ch = SPACE;
					Nonblanks = TRUE;
					break;

				case 4 :
					ch = SPACE;
					break;

				case 5 :
					if ( OutBufferIndex ) {
						if ( CountNonblanks( OutBuffer ) )
							Nonblanks = TRUE;

						if ( Nonblanks || IncludeBlankLines ) {
							QPuts( OutBuffer );
							OutBufferIndex = 0;
							OutBuffer[OutBufferIndex] = '\0';
							Dirty = TRUE;
						}

						if ( Dirty )
							if ( CL_EndLine() == USER_ABORT )
								goto UserAbort;
					}

					QPuts( L"    \u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014\u2014" );
					if ( CL_EndLine( 2 ) == USER_ABORT )
						goto UserAbort;

					ch			= SPACE;
					Dirty		= FALSE;
					Nonblanks	= FALSE;
					break;
				}
			}

			continue;
		}

		if ( InTag ) {
			if ( ( InTag == 1 ) && ( ch == '-' ) ) {
				if ( wcscmp( TagName, L"!-" ) == 0 ) {
					InTag = 0;
					InComment = TRUE;
					j = 0;
					TagName[j] = TagAttributes[j] = '\0';
					continue;
				}
			}

			if ( ( InTag == 1 ) && j && ( iswspace( ch ) || ( ch == '/' ) ) ) {
				InTag = 2;
				j = 0;
			}

			if ( InTag == 1 ) {
				if ( j < ( SHORT_BUF_LEN - 1 ) ) {
					TagName[j++] = towupper( ch );
					TagName[j] = '\0';
				}
			}
			else if ( ( InTag == 2 ) && ( j || ( iswspace( ch ) == 0 ) ) ) {
				if ( j < ( SHORT_BUF_LEN - 1 ) ) {
					TagAttributes[j++] = ch;
					TagAttributes[j] = '\0';
				}
			}

			continue;
		}

		if ( InHead && !InTitle )
			continue;

		if ( BodyOnly )
			if ( !InBody && !InTitle )
				continue;

		if ( InScript || InApplet || InStyle )
			continue;

		if ( InSelect == 2 )
			continue;

		if ( ( ch == '&' ) && !iswspace( MMPeekAhead( &MMFile, 0 ) ) ) {
			TagAttributes[0] = ch;
			TagAttributes[1] = '\0';
			int ahead = 0;

			for ( int i = 0; i < 64; i++ ) {
				ahead = MMPeekAhead( &MMFile, i );
				
				if ( ahead <= 0 )
					break;

				TagAttributes[i + 1] = ahead;
				TagAttributes[i + 2] = '\0';

				if ( ahead == ';' )
					break;
			}

			if ( ahead == ';' ) {
				int test = ParseEntity( TagAttributes, NULL );
				TagAttributes[0] = '\0';

				if ( test > 0 ) {
					ch = test;

					do {
						ahead = MMGetChar( &MMFile );
					} while ( ahead != ';' );
				}
			}
		}

		if ( iswspace( ch ) && !InPreBlock ) {
			ch = SPACE;
			if ( lastch == SPACE )
				continue;
		}

		
		if ( PrintLongEntity( ch, OutBuffer, &OutBufferIndex ) )
			continue;
		else {
			if ( OutBufferIndex >= ( MAX_ARG_LEN - 16 ) ) {
				if ( CountNonblanks( OutBuffer ) )
					Nonblanks = TRUE;

				QPuts( OutBuffer );
				OutBufferIndex = 0;
				OutBuffer[OutBufferIndex] = '\0';
				Dirty = TRUE;
			}

			if ( ch > 0xffff ) {
				MM_CreateSurrogatePair( ch, OutBuffer + OutBufferIndex );
				OutBufferIndex += 2;
				OutBuffer[OutBufferIndex] = '\0';
			}
			else {
				OutBuffer[OutBufferIndex++] = ch;
				OutBuffer[OutBufferIndex] = '\0';
			}
		}
	}

	if ( OutBufferIndex ) {
		QPuts( OutBuffer );
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}
	
	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return USER_ABORT;
}



DLLExports int WINAPI dehtml( LPTSTR inString )
{
	if ( !inString )
		return 1;


	int Flags				= 2;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	BOOL WriteBOM			= TRUE;
	int rv					= 0;
	int fhrv				= 0;

	_DeHtmlArgs args		= { 0, 0 };


	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( DeHTMLHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( DEHTML_COMMAND_VER ) );
			}
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(args.CodePage), &(args.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/B" ) )
				Flags |= SH_BODY_ONLY;
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				Flags |= SH_COMMENTS;
			else if ( !_wcsicmp( arg[i], L"/DE" ) )
				return DH_DumpEntities();
			else if ( !_wcsicmp( arg[i], L"/E" ) )
				Flags |= SH_STRIP_BLANK_LINES;
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags |= SH_HEADER;
			else if ( !_wcsicmp( arg[i], L"/M" ) )
				Flags |= SH_CHECK_META_TAGS;
			else if ( !_wcsicmp( arg[i], L"/N" ) )
				Flags |= SH_NOSCRIPT;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/O", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				unsigned int Value = 0;

				if ( ParseInt( arg[i] + j, &Value, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				if ( Value > 2 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Value must be 0 - 2\r\n" );
					goto ErrorExit;
				}

				Flags = ( Flags & ~SH_OPTIONS ) | ( 2 - Value );
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				Flags |= SH_REMOVE_TITLE;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}

		}
	}


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	if ( !StdinSpecified ) {
		if ( ( ConsoleHandle( STD_INPUT_HANDLE ) == 0 ) || !FilenameSpecified ) {
			fhrv = StripHTML( (LPTSTR) StdinFilename, Flags, &args );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &args, StripHTML, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit :

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	FH_FreeLinkedList( &FirstRange );
	return 3;
}



DLLExports int WINAPI f_mkentities( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	if ( wcslen( String ) >= PLUGIN_BUFFER_MAX ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"String too long!  %u characters max.\r\n", PLUGIN_BUFFER_MAX - 1 );
		return -1;
	}


	const int MiniBufLen				= 16;

	wchar_t temp[PLUGIN_BUFFER_MAX]		= L"";
	wchar_t tmp2[MiniBufLen]			= L"";

	size_t j							= 0;
	unsigned int HighChar				= 0;


	for ( wchar_t *ch = String; *ch; ch++ ) {

		if ( *ch == '"' )
			wcscpy_s( tmp2, MiniBufLen, L"&quot;" );
		else if ( *ch == '%' )
			wcscpy_s( tmp2, MiniBufLen, L"&37;" );
		else if ( *ch == '&' )
			wcscpy_s( tmp2, MiniBufLen, L"&amp;" );
		else if ( *ch == '<' )
			wcscpy_s( tmp2, MiniBufLen, L"&lt;" );
		else if ( *ch == '>' )
			wcscpy_s( tmp2, MiniBufLen, L"&gt;" );
		else if ( ( *ch >= 0x0080 ) && ( *ch <= 0x009f ) )
			wsprintf( tmp2, L"&#x%x;", Windows1252[*ch & 0x001f] );
		else if ( iswcntrl( *ch ) )
			wsprintf( tmp2, L"&#%u;", *ch );
		else if ( ( ( ch[0] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_HIGH ) && ( ( ch[1] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_LOW ) ) {
			HighChar = ( ( ( ch[0] & SURROGATE_MASK ) << 10 ) | ( ch[1] & SURROGATE_MASK ) ) + SURROGATE_OFFSET;
			if ( HighChar > UNICODE_MAX_CHAR )
				HighChar = 0xfffd;

			wsprintf( tmp2, L"&#x%x;", HighChar );
			ch++;
		}
		else if ( j < ( PLUGIN_BUFFER_MAX - 1 ) ) {
			temp[j++]	= *ch;
			temp[j]		= '\0';
		}
		else
			goto Overflow;

		if ( tmp2[0] ) {
			if ( ( j + wcslen( tmp2 ) ) >= PLUGIN_BUFFER_MAX )
				goto Overflow;
			else {
				j += wcslen( tmp2 );
				wcscat_s( temp, PLUGIN_BUFFER_MAX, tmp2 );
				tmp2[0] = '\0';
			}
		}
	}


	wcscpy_s( String, PLUGIN_BUFFER_MAX, temp );
	return 0;

Overflow:

	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Buffer overflow!  Resulting string exceeds %u characters.\r\n", PLUGIN_BUFFER_MAX - 1 );
	return -1;
}


#endif	//	DEHTML_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-28
				Now uses HtmlEntities.cpp.  </TH> adds a tab.  Added /R to
				remove titles.  Added a version number to this source file.
				Help text tweaks.  /?? now shows this file's version info.

	1.0.1.0		2023-09-28
				StripHTML() filename is defined as a constant:  fix for mmfiles 1.4.4.2+

	1.0.2.0		2023-10-17
				Whitespace is not smashed inside <PRE> blocks.

	1.0.2.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
