
//			@INIVALUE function, and _INIVALUERC variable:



#define IV_FILE_ERRORS				0x0001
#define IV_SECTION_WILDCARDS		0x0002
#define IV_ENTRY_WILDCARDS			0x0004
#define IV_DEBUG					0x1000



wchar_t IniValueExit[64] = L"";



BOOL IsSectionHeader( LPTSTR inString, LPTSTR outSectionName )
{
	if ( !inString )
		return FALSE;
	else if ( inString[0] == '\0' )
		return FALSE;


	int FirstNameChar = -1, LastNameChar = -1;
	BOOL OpenBracket = FALSE, CloseBracket = FALSE;
	wchar_t ch = '\0';

	for ( int i = 0; inString[i]; i++ ) {
		ch = inString[i];

		if ( !OpenBracket && ( ch == '[' ) )
			OpenBracket = TRUE;
		else if ( OpenBracket && !CloseBracket && ( ch == ']' ) )
			CloseBracket = TRUE;
		else if ( !iswspace( ch ) && ( !OpenBracket || CloseBracket ) )
			return FALSE;
		else if ( !iswspace( ch ) ) {
			LastNameChar = i;

			if ( FirstNameChar == -1 )
				FirstNameChar = i;
		}
	}

	if ( !OpenBracket || !CloseBracket || ( FirstNameChar == -1 ) )
		return FALSE;


	if ( !outSectionName )
		return TRUE;

	int j = 0;
	for ( int i = FirstNameChar; i <= LastNameChar; i++ ) {
		if ( j < ( MAX_ARG_LEN - 1 ) )
			outSectionName[j++] = towupper( inString[i] );
	}

	outSectionName[j] = '\0';
	return TRUE;
}


BOOL IsIniEntry( LPTSTR inString, LPTSTR outName, LPTSTR outValue )
{
	if ( !inString )
		return FALSE;
	else if ( inString[0] == '\0' )
		return FALSE;


	int FirstNameChar = -1, LastNameChar = -1, EqualsSign = -1;
	int FirstValueChar = -1, LastValueChar = -1;
	wchar_t ch = '\0';


	for ( int i = 0; inString[i]; i++ ) {
		ch = inString[i];

		if ( ( ch == '=' ) && ( EqualsSign == -1 ) )
			EqualsSign = i;
		else if ( iswspace( ch ) == 0 ) {

			if ( EqualsSign == -1 ) {
				LastNameChar = i;

				if ( FirstNameChar == -1 )
					FirstNameChar = i;
			}
			else {
				LastValueChar = i;

				if ( FirstValueChar == -1 )
					FirstValueChar = i;
			}
		}
	}

	if ( ( EqualsSign == -1 ) || ( FirstNameChar == -1 ) )
		return FALSE;

	int j = 0;

	if ( outName != NULL ) {
		for ( int i = FirstNameChar; i <= LastNameChar; i++ ) {
			if ( j < ( MAX_ARG_LEN - 1 ) )
				outName[j++] = towupper( inString[i] );
		}

		outName[j] = '\0';
	}

	j = 0;
	if ( outValue )
		outValue[j] = '\0';

	if ( FirstValueChar == -1 )
		return TRUE;

	if ( ( LastValueChar > FirstValueChar ) && ( ( inString[FirstValueChar] == '\'' ) || ( inString[FirstValueChar] == '"' ) ) && ( inString[FirstValueChar] == inString[LastValueChar] ) ) {
		if ( LastValueChar == ( FirstValueChar + 1 ) )
			return TRUE;

		FirstValueChar++;
		LastValueChar--;
	}

	if ( outValue != NULL ) {
		for ( int i = FirstValueChar; i <= LastValueChar; i++ ) {
			if ( j < ( MAX_ARG_LEN - 1 ) )
				outValue[j++] = inString[i];
		}

		outValue[j] = '\0';
	}
	
	return TRUE;
}



DLLExports int WINAPI f_inivalue( LPTSTR String )
{
	if ( !String )
		return 1;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	_MMFile MMFile;
	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	wchar_t SectionName[MAX_ARG_LEN]		= L"";
	wchar_t EntryName[MAX_ARG_LEN]			= L"";
	unsigned int Index = 0, Flags = 0;

	wchar_t LineBuffer[MAX_ARG_LEN]			= L"";
	wchar_t CurrentSection[MAX_ARG_LEN]		= L"";
	wchar_t CurrentEntry[MAX_ARG_LEN]		= L"";
	wchar_t CurrentValue[MAX_ARG_LEN]		= L"";
	BOOL RightSection = FALSE, AllSections = FALSE, FoundSection = FALSE;
	BOOL FoundEntry = FALSE;
	unsigned int Count = 0;
	int LineNumber = -1;
	int rv = 0;

	BOOL FileErrors = FALSE, SectionWildcards = FALSE, EntryWildcards = FALSE, Debug = FALSE;


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		wcscpy_s( IniValueExit, 64, L"Syntax error" );
		return -1;
	}

	if ( !FH_PathPart( arg[0], NULL ) ) {
		GetWindowsDirectory( Filename, MAX_FILENAME_LEN );

		if ( wcslen( Filename ) )
			if ( Filename[(int) wcslen( Filename ) - 1] != '\\' )
				wcscat_s( Filename, MAX_FILENAME_LEN, L"\\" );

		wcscat_s( Filename, MAX_FILENAME_LEN, arg[0] );
	}
	else
		FH_Truename( arg[0], Filename );


	if ( arglen[5] ) {
		if ( ParseInt( arg[5], &Flags, NULL ) ) {
			DisplayErrorHeader();
			_wcsupr_s( arg[5], MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad flags: \"%s\"\r\n", arg[5] );
			wcscpy_s( IniValueExit, 64, L"Syntax error" );
			return -1;
		}
	}

	FileErrors			= ( Flags & IV_FILE_ERRORS ) > 0;
	SectionWildcards	= ( Flags & IV_SECTION_WILDCARDS ) > 0;
	EntryWildcards		= ( Flags & IV_ENTRY_WILDCARDS ) > 0;
	Debug				= ( Flags & IV_DEBUG ) > 0;


	if ( arglen[1] ) {
		if ( wcscmp( arg[1], L"*" ) == 0 )
			FoundSection = AllSections = TRUE;
		else {
			if ( ! SectionWildcards ) {
				if ( ( wcschr( arg[1], '[' ) != NULL ) || ( wcschr( arg[1], ']' ) ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[1], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad section name: \"%s\"\r\n", arg[1] );
					wcscpy_s( IniValueExit, 64, L"Syntax error" );
					return -1;
				}
			}

			wcscpy_s( SectionName, MAX_ARG_LEN, arg[1] );
		}
	}
	else if ( SectionWildcards ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing section name\r\n" );
		wcscpy_s( IniValueExit, 64, L"Syntax error" );
		return -1;
	}
	else
		RightSection = TRUE;


	if ( arglen[2] ) {
		if ( wcschr( arg[2], '=' ) != NULL ) {
			DisplayErrorHeader();
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad entry name: \"%s\"\r\n", arg[2] );
			wcscpy_s( IniValueExit, 64, L"Syntax error" );
			return -1;
		}

		wcscpy_s( EntryName, MAX_ARG_LEN, arg[2] );
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing entry name\r\n" );
		wcscpy_s( IniValueExit, 64, L"Syntax error" );
		return -1;
	}


	if ( arglen[3] ) {
		if ( arg[3][0] == '-' )
			Index = UINT_MAX;
		else {
			if ( ParseInt( arg[3], &Index, NULL ) ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[3], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Bad index: \"%s\"\r\n", arg[3] );
				wcscpy_s( IniValueExit, 64, L"Syntax error" );
				return -1;
			}
		}
	}


	if ( arglen[4] )
		wcscpy_s( String, MAX_ARG_LEN, arg[4] );
	else
		String[0] = '\0';


	rv = OpenMMFile( &MMFile, Filename, FileErrors );

	if ( rv ) {

		if ( rv == ERROR_EMPTY_FILE )
			wcscpy_s( IniValueExit, 64, L"File empty" );
		else
			wsprintf( IniValueExit, L"File error %d", MMFile.ErrorCode );

		if ( FileErrors ) {
			if ( rv == ERROR_EMPTY_FILE ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"File is empty.\r\n" );
			}
			return -1;
		}
		else {
			if ( Index == UINT_MAX )
				wcscpy_s( String, MAX_ARG_LEN, L"0" );

			return 0;
		}
	}


	wchar_t *ch = NULL;


	do {
		if ( MMGetLine( &MMFile, LineBuffer, MAX_ARG_LEN, TRUE ) )
			break;

		LineNumber++;
		ch = LineBuffer;
		while ( iswspace( *ch ) )
			ch++;

		if ( ( *ch == '\0' ) || ( *ch == ';' ) )
			continue;

		if ( IsSectionHeader( LineBuffer, CurrentSection ) ) {
			if ( Debug )
				Printf( L"   * Found section \"%s\"\r\n", CurrentSection );

			if ( SectionWildcards )
				RightSection = ( WildcardComparison( SectionName, CurrentSection, 0, 1 ) == 0 );
			else
				RightSection = ( _wcsicmp( CurrentSection, SectionName ) == 0 );

			if ( RightSection )
				FoundSection = TRUE;

			continue;
		}

		if ( !RightSection && !AllSections )
			continue;

		if ( !IsIniEntry( LineBuffer, CurrentEntry, CurrentValue ) )
			continue;

		if ( Debug )
			Printf( L"   * Found entry \"%s\" with value \"%s\"\r\n", CurrentEntry, CurrentValue );

		if ( !EntryWildcards && _wcsicmp( CurrentEntry, EntryName ) )
			continue;

		if ( EntryWildcards && WildcardComparison( EntryName, CurrentEntry, 0, 1 ) )
			continue;

		if ( ( Index != UINT_MAX ) && ( Count == Index ) ) {
			wcscpy_s( String, MAX_ARG_LEN, CurrentValue );
			wsprintf( IniValueExit, L"Found %d", LineNumber );
			FoundEntry = TRUE;
			break;
		}

		Count++;

	} while ( TRUE );
	
	CloseMMFile( &MMFile );


	if ( Index == UINT_MAX ) {
		wsprintf( String, L"%u", Count );
		wsprintf( IniValueExit, L"Count %d", Count );
	}
	else {
		if ( !FoundSection )
			wcscpy_s( IniValueExit, 64, L"No section" );
		else if ( !FoundEntry )
			wsprintf( IniValueExit, L"No entry %d", Index );
	}
	
	return 0;
}



DLLExports int WINAPI _inivaluerc( LPTSTR outString )
{
	if ( !outString )
		return 1;

	wcscpy_s( outString, MAX_ARG_LEN, IniValueExit );
	return 0;
}
