
//			PARSEARGS command:


#ifndef PARSEARGS_COMMAND_VER
#define PARSEARGS_COMMAND_VER			1.0.0.0


#define PA_MAX_LINE_LEN			( MAX_ARG_LEN + 64 )



const wchar_t ParseArgsHelpText[] = L"Divide a string into arguments.\n\
\n\
PARSEARGS /A:array /F:flags /Q /V:var !string\r\n\
\n\
\t/A:  name of array to receive the arguments; defaults to ARG \n\
\t/F:  parse flags:\n\
\t      1: divide args at spaces\n\
\t      2: divide args at commas\n\
\t      4: slashes kludge: treat /A/B like /A /B\n\
\t      8: quotes kludge: treat /A\"foo\" like /A:\"foo\"\n\
\t     16: divide arg at first unquoted equals sign\n\
\t     64: retain double quotes\n\
\t/Q   quietly\n\
\t/V:  name of an environment variable containing a string\n\
\n\
In the flags argument, you should specify at least one of\n\
1, 2, or 16; you may specify more than one.  If you don\u2019t\n\
specify any, then 1 is assumed.\n\
\n\
You may pass the string to parse either on the command line\n\
following an exclamation point, or else in an environment\n\
variable named with /V:.\n";




DLLExports int WINAPI parseargs( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t ArrayName[SHORT_BUF_LEN]	= L"";
	wchar_t VarName[SHORT_BUF_LEN]		= L"";
	wchar_t VarValue[MAX_ARG_LEN]		= L"";
	wchar_t CmdBuf[PA_MAX_LINE_LEN]		= L"";
	LPTSTR String						= NULL;

	unsigned int Flags					= 0;
	BOOL Quiet							= FALSE;
	int rv								= 0;


	String = wcschr( inString, '!' );

	if ( String ) {
		*String = '\0';
		String++;
	}


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( ParseArgsHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( PARSEARGS_COMMAND_VER ) );
			else if ( !_wcsnicmp( arg[i], L"/A", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( !GoodArrayName( arg[i] + j ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad array name: \"%s\"\r\n", arg[i] + j  );
					return 1;
				}

				wcscpy_s( ArrayName, SHORT_BUF_LEN, arg[i] + j );
				_wcsupr_s( ArrayName, SHORT_BUF_LEN );
			}
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				Flags |= PARSE_BREAK_COMMAS;
			else if ( !_wcsnicmp( arg[i], L"/F", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				unsigned int TempFlags = 0;

				if ( ParseInt( arg[i] + j, &TempFlags, NULL ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad flags: \"%s\"\r\n", arg[i] );
					return 1;
				}

				Flags |= TempFlags;
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Quiet = TRUE;
			else if ( !_wcsicmp( arg[i], L"/U" ) )
				Flags |= PARSE_FORCE_UPPERCASE;
			else if ( !_wcsicmp( arg[i], L"/V" ) )
				wcscpy_s( VarName, SHORT_BUF_LEN, L"ARGS" );
			else if ( !_wcsnicmp( arg[i], L"/V", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( !GoodArrayName( arg[i] + j ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad variable name: \"%s\"\r\n", arg[i] + j  );
					return 1;
				}

				wcscpy_s( VarName, SHORT_BUF_LEN, arg[i] + j );
				_wcsupr_s( VarName, SHORT_BUF_LEN );
			}
			else if ( !_wcsicmp( arg[i], L"/W" ) )
				Flags |= PARSE_BREAK_SPACES;
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}

	if ( !String && ( VarName[0] == '\0' ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Must specify either !string or /V:\r\n" );
		return 1;
	}
	else if ( String && VarName[0] ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Cannot combine !string and /V:\r\n" );
		return 1;
	}

	if ( ArrayName[0] == '\0' )
		wcscpy_s( ArrayName, SHORT_BUF_LEN, L"ARG" );

	if ( ( Flags & ( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_EQUALS_KLUDGE ) ) == 0 )
		Flags |= PARSE_BREAK_SPACES;

	
	if ( VarName[0] ) {
		GetEnvironmentVariable( VarName, VarValue, MAX_ARG_LEN );

		if ( VarValue[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Environment variable not found: \"%s\"\r\n", VarName );
			return 2;
		}

		rv = ParseArgs( VarValue, Flags );
	}
	else
		rv = ParseArgs( String, Flags );

	if ( rv < 0 )
		return 2;


	wcscpy_s( CmdBuf, PA_MAX_LINE_LEN, L" /Q " );				//  delete any existing array
	wcscat_s( CmdBuf, PA_MAX_LINE_LEN, ArrayName );
	CallInternal( L"unsetarray", CmdBuf );

	if ( numargs ) {											//  and create it anew
		wsprintf( CmdBuf, L" %s[%d]", ArrayName, numargs );
		CallInternal( L"setarray", CmdBuf );
	}


	if ( !Quiet )
		Printf( L"\r\n%d args found:\r\n\n", numargs );

	for ( int i = 0; i < numargs; i++ ) {
		if ( !Quiet )
			Printf( L"%s[%d] = %s\r\n", ArrayName, i, arg[i] );

		wsprintf( CmdBuf, L" %s[%d]=%s", ArrayName, i, arg[i] );
		CallInternal( L"set", CmdBuf );
	}

	wsprintf( CmdBuf, L" %s_N=%d", ArrayName, numargs );
	CallInternal( L"set", CmdBuf );
	return 0;
}


#endif	//	PARSEARGS_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-27
				Adds a version number to this source file.  Help text tweaks.

*/
