
//			TEXT2CLIP and CLIP2TEXT commands:


#ifndef TEXT2CLIP_VERSION
#define TEXT2CLIP_VERSION			1.0.0.1


#define T2C_FLAG_ERROR_MESSAGES		0x0001
#define T2C_FLAG_FIX_QUOTES			0x0002
#define T2C_FLAG_APPEND				0x0004
#define T2C_FLAG_DEBUG				0x0100

#define C2T_BUFFER_SIZE				8193



const wchar_t Text2ClipHelpText[]	= L"Copy text from a file onto the clipboard.\n\
\n\
TEXT2CLIP /A /CP:n /Q /T filename \r\n\
\n\
\t/A    append to the clipboard\n\
\t/CP:  interpret OEM text as code page n\n\
\t/Q    replace ASCII quotes and apostrophes\n\
\t/T    quietly\n\
\n\
Only one filename is allowed.  Text may be piped or redirected\n\
into TEXT2CLIP.\n";


const wchar_t Clip2TextHelpText[]	= L"Copy text from the clipboard to a file or stdout.\n\
\n\
CLIP2TEXT /A /O /P /T /UTF8 /UTF16 filename \r\n\
\n\
\t/A      apppend to an existing file\n\
\t/O      overwrite an existing file\n\
\t/P      page output\n\
\t/T      quietly\n\
\t/UTF8   write file as UTF-8\n\
\t/UTF16  write file as UTF-16 (default)\n\
\n\
Only one filename is allowed.  If no filename is specified, the\n\
default is to dump the clipboard to stdout.\n";





int MyEmptyClipboard()
{
	int rv			= 0;


	if ( OpenClipboard( GetConsoleWindow() ) ) {

		if ( EmptyClipboard() )
			rv = 0;
		else
			rv = GetLastError();

		CloseClipboard();
	}
	else
		rv = GetLastError();

	return rv;
}



int FileToClipboard( LPTSTR Filename, unsigned int Flags, unsigned int CodePage, unsigned int EbcdicCodePage, unsigned __int64 *outChars )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t AposWord[SHORT_BUF_LEN]		= L"";

	_MMFile MMFile;

	size_t nEOLs			= 0;
	size_t nHighChars		= 0;
	size_t nOtherChars		= 0;
	size_t BufferSize		= 0;
	HGLOBAL MemHandle		= NULL;
	wchar_t *Buffer			= NULL;
	int ch					= 0;
	int lastch				= 0;
	int nextch				= 0;
	size_t i				= 0;

	LPTSTR OldClipboardText		= NULL;
	size_t OldClipboardLen		= 0;
	size_t OldClipboardChars	= 0;

	HWND WindowHandle			= GetConsoleWindow();
	BOOL ReadFail				= FALSE;
	BOOL Debug					= ( Flags & T2C_FLAG_DEBUG ) > 0;
	int gurv					= 0;
	int rv						= 0;



	if ( Debug )
		_cputws( L"   A:  About to open file\n" );

	if ( OpenMMFile( &MMFile, Filename, ( Flags & T2C_FLAG_ERROR_MESSAGES ) | MM_FLAG_ALLOW_EBCDIC | MM_FLAG_HIGH_CHARS, CodePage, EbcdicCodePage ) ) {
		if ( MMFile.ErrorCode )
			return MMFile.ErrorCode;
		else
			return -2;
	}


	if ( Debug )
		_cwprintf( L"   B:  FileEncoding = %u\n", MMFile.FileEncoding );


	//		count characters in the file:

	while ( MMFile.Index < MMFile.FileSize ) {

		ch = MMGetChar( &MMFile );

		if ( ch < 0 ) {
			ReadFail = TRUE;
			break;
		}
		else if ( ch == 0 )
			;
		else if ( ch > 0xffff )
			nHighChars++;
		else if ( _MMIsEOL( ch ) ) {
			nEOLs++;
			if ( MMPeekAhead( &MMFile, 0 ) == _MM2PartEOL( ch ) )
				MMGetChar( &MMFile );
		}
		else
			nOtherChars++;
	}

	if ( ReadFail ) {
		CloseMMFile( &MMFile );
		return -3;
	}

	
	if ( Flags & T2C_FLAG_APPEND ) {
		OldClipboardText = GetClipboardText( &OldClipboardLen );

		if ( OldClipboardText )
			OldClipboardChars = MM_CodePointsIn( OldClipboardText );
	}


	BufferSize = ( 2 * nEOLs ) + ( 2 * nHighChars ) + nOtherChars + OldClipboardLen + 17;

	if ( Debug )
		_cwprintf( L"   C:  nEOLs = %I64u   nHighChars = %I64u   nOtherChars = %I64u   OldClipboardLen = %I64u   BufferSize = %I64u\n", (unsigned __int64) nEOLs, (unsigned __int64) nHighChars, (unsigned __int64) nOtherChars, (unsigned __int64) OldClipboardLen, (unsigned __int64) BufferSize );


	MMFile.Index = MMFile.BOMSize;
	ReadFail = TRUE;

	MemHandle = GlobalAlloc( GHND, BufferSize * sizeof( wchar_t ) );
	if ( MemHandle ) {

		if ( Debug )
			_cputws( L"   D:  Buffer allocated.\n" );

		Buffer = (wchar_t*) GlobalLock( MemHandle );
		if ( Buffer ) {

			if ( Debug )
				_cputws( L"   E:  Buffer locked.\n" );

			if ( OldClipboardText ) {
				memcpy( Buffer, OldClipboardText, OldClipboardLen * sizeof( wchar_t ) );
				free( OldClipboardText );
				OldClipboardText = NULL;
				i = OldClipboardLen;
			}

			ReadFail = FALSE;
			
			while ( MMFile.Index < MMFile.FileSize ) {

				ch = MMGetChar( &MMFile );

				if ( ch < 0 ) {
					ReadFail = TRUE;
					break;
				}
				else if ( ch == 0 )
					;
				else if ( _MMIsEOL( ch ) ) {
					Buffer[i++] = '\r';
					Buffer[i++] = '\n';

					if ( MMPeekAhead( &MMFile, 0 ) == _MM2PartEOL( ch ) )
						MMGetChar( &MMFile );

					lastch = '\n';
				}
				else if ( ch > 0xffff ) {
					MM_CreateSurrogatePair( ch, Buffer + i );
					i += 2;
					lastch = ch;
				}
				else {

					if ( Flags & T2C_FLAG_FIX_QUOTES ) {
						nextch = MMPeekAhead( &MMFile, 0 );

						if ( ReplaceDoubles && ( ch == APOSTROPHE ) && ( nextch == APOSTROPHE ) ) {
							if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( MMPeekAhead( &MMFile, 1 ) ) ) {
								ch = QUOTE;
								MMGetChar( &MMFile );
								nextch = MMPeekAhead( &MMFile, 0 );
							}
						}

						if ( ReplaceDoubles && ReplaceGraves && ( ch == GRAVE ) && ( nextch == GRAVE ) ) {
							if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( MMPeekAhead( &MMFile, 1 ) ) ) {
								ch = LDQUOTE;
								MMGetChar( &MMFile );
								nextch = MMPeekAhead( &MMFile, 0 );
							}
						}

						if ( ReplaceGraves && ( ch == GRAVE ) )
							if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( MMPeekAhead( &MMFile, 1 ) ) )
								ch = LSQUOTE;

						if ( ( ch == QUOTE ) || ( ch == FWQUOTE ) ) {
							if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswspace( nextch ) || _MMIsEOL( nextch ) ) )
								ch = DoublePrime;
							else if ( iswalpha( lastch ) && !iswalpha( nextch ) )
								ch = CloseQuote;
							else if ( iswalpha( nextch ) && !iswalpha( lastch ) )
								ch = OpenQuote;
							else if ( iswgraph( lastch ) && ( lastch != OpenSQuote ) )
								ch = CloseQuote;
							else {
								ch = OpenQuote;

								if ( iswspace( lastch ) )
									if ( _MMIsEOL( nextch ) )
										ch = CloseQuote;
							}
						}
						else if ( ( ch == APOSTROPHE ) || ( ch == FWAPOSTROPHE ) ) {
							if ( iswgraph( lastch ) && ( lastch != OpenQuote ) ) {
								ch = CloseSQuote;

								if ( iswalnum( lastch ) && iswalpha( nextch ) )
									ch = Apostrophe;
							}
							else {
								ch = OpenSQuote;

								if ( iswspace( lastch ) )
									if ( _MMIsEOL( nextch ) )
										ch = CloseSQuote;
							}

							if ( !iswalnum( lastch ) && iswalpha( nextch ) ) {
								//		check for word beginning with apostrophe ....

								AposWord[0] = APOSTROPHE;
								int peek = 0;
								for ( int j = 0; j < 15; j++ ) {
									peek = MMPeekAhead( &MMFile, j );
									AposWord[j + 1] = ( peek < SPACE ) ? SPACE : peek;
								}

								AposWord[16] = 0x00;
								if ( IsWordBeginningWithApostrophe( AposWord ) )
									ch = Apostrophe;
							}
							else if ( IsVowel2( lastch ) && IsVowel2( nextch ) )
								ch = Okina;
					
							if ( ( iswalnum( lastch ) == 0 ) && iswdigit( nextch ) )
								if ( iswdigit( MMPeekAhead( &MMFile, 1 ) ) )
									if ( iswalnum( MMPeekAhead( &MMFile, 2 ) ) == 0 )
										ch = Apostrophe;

							if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswdigit( nextch ) || IsFraction( nextch ) || iswspace( nextch ) || _MMIsEOL( nextch ) ) )
								ch = Prime;
						}
						else if ( ( ch == '-' ) || ( ch == 0xff0d ) ) {
							if ( ( ( lastch != '-' ) && ( lastch != 0xff0d ) ) && ( nextch == ch ) ) {
								int tch = MMPeekAhead( &MMFile, 1 );
								if ( ( tch != '-' ) && ( tch != 0xff0d ) ) {
									ch =EmDash;
									MMGetChar( &MMFile );
								}
							}
						}
					}

					lastch = ch;
					Buffer[i++] = ch;
				}
			}

			Buffer[i] = '\0';
			gurv = GlobalUnlock( MemHandle );

			if ( Debug )
				_cwprintf( L"   F:  GlobalUnlock() returned %d; GetLastError() returns %u.\n", gurv, GetLastError() );
		}
	}

	if ( Debug )
		_cputws( L"   G:  About to close MMFile.\n" );

	CloseMMFile( &MMFile );

	if ( Debug )
		_cputws( L"   H:  MMFile closed.\n" );


	if ( ReadFail ) {
		GlobalFree( MemHandle );
		return -4;
	}


	if ( Debug )
		_cwprintf( L"   I:  About to OpenClipboard( 0x%04X ).\n", WindowHandle );

	if ( OpenClipboard( GetConsoleWindow() ) ) {

		if ( Debug )
			_cputws( L"   J:  Clipboard opened.\n" );

		if ( EmptyClipboard() ) {

			if ( Debug )
				_cputws( L"   K:  Clipboard emptied.\n" );

			if ( SetClipboardData( CF_UNICODETEXT, MemHandle ) ) {
				rv = 0;

				if ( Debug )
					_cputws( L"   L:  SetClipboardData() succeeded.\n" );
			}
			else {
				rv = GetLastError();

				if ( Debug )
					_cwprintf( L"   M:  SetClipboardData() failed; rv = %d\n", rv );

				if ( rv == ERROR_INVALID_HANDLE ) {

					if ( Debug )
						_cputws( L"   M1:  Retry...\n" );

					Sleep( 150 );
					if ( SetClipboardData( CF_UNICODETEXT, MemHandle ) ) {
						rv = 0;

						if ( Debug )
							_cputws( L"   M2:  Second call to SetClipboardData() succeeded.\n" );
					}
					else {
						rv = GetLastError();

						if ( Debug )
							_cwprintf( L"   M3:  Second call to SetClipboardData() failed; rv = %d\n", rv );
					}
				}

				if ( rv )
					GlobalFree( MemHandle );
			}
		}
		else
			rv = GetLastError();

		if ( Debug )
			_cputws( L"   N:  About to close clipboard.\n" );

		CloseClipboard();
	}
	else
		rv = GetLastError();

	if ( Debug )
		_cputws( L"   O:  About to set *outChars.\n" );

	if ( !rv && outChars )
		*outChars = ( 2 * nEOLs ) + nHighChars + nOtherChars;

	if ( Debug )
		_cwprintf( L"   Z:  Exit with return code %d.\n", rv );

	return rv;
}



DLLExports int WINAPI text2clip( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";
	int Flags								= 0;
	BOOL Empty								= FALSE;
	BOOL Quiet								= FALSE;
	unsigned int CodePage					= 0;
	unsigned int EbcdicCodePage				= 0;
	unsigned __int64 CharsWritten			= 0;
	int fhrv								= 0;



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( Text2ClipHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( TEXT2CLIP_VERSION ) );
			else if ( !wcscmp( arg[i], L"/$" ) )
				Flags |= T2C_FLAG_DEBUG;
			else if ( !_wcsicmp( arg[i], L"/A" ) )
				Flags |= T2C_FLAG_APPEND;
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &CodePage, &EbcdicCodePage ) )
					return 1;
			}
			else if ( !_wcsicmp( arg[i], L"/E" ) )
				Empty = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Flags |= T2C_FLAG_FIX_QUOTES;
			else if ( !_wcsicmp( arg[i], L"/T" ) )
				Quiet = TRUE;
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( Filename[0] == '\0' ) {
				if ( QueryIsURL( arg[0] ) )
					wcscpy_s( Filename, MAX_FILENAME_LEN, arg[0] );
				else
					FH_Truename( arg[i], Filename );
			}
			else {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}


	if ( Empty && ( Filename[0] == '\0' ) ) {
		int rv		= MyEmptyClipboard();

		if ( rv ) {
			GetErrorText( rv, ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"%s error emptying clipboard.\r\n" );
			return 2;
		}

		if ( !Quiet )
			Printf( L"Clipboard emptied.\r\n" );

		return 0;
	}


	if ( ( Filename[0] == '\0' ) && ConsoleHandle( STD_INPUT_HANDLE ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No filename specified\r\n" );
		return 1;
	}

	if ( !_wcsicmp( Filename, ClipFilename ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"CLIP: is not a valid filename\r\n" );
		return 1;
	}


	if ( Flags & T2C_FLAG_FIX_QUOTES )
		InitializeFixQuotesVars();


	if ( ( Filename[0] == '\0' ) && !ConsoleHandle( STD_INPUT_HANDLE ) ) {
		fhrv = FileToClipboard( (LPTSTR) StdinFilename, Flags, CodePage, EbcdicCodePage, &CharsWritten );
		if ( fhrv )
			goto ErrorExit;
	}

	if ( Filename[0] ) {
		fhrv = FileToClipboard( Filename, Flags, CodePage, EbcdicCodePage, &CharsWritten );
		if ( fhrv )
			goto ErrorExit;
	}


	if ( !Quiet )
		Printf( L"Copied %s characters to the clipboard.\r\n", Commafy( CharsWritten ) );

	return 0;


ErrorExit :

	DisplayErrorHeader();
	MM_GetErrorText( fhrv, ErrorText );
	Qprintf( ERRHANDLE, L"Error copying file \"%s\" to the clipboard:\r\n", Filename[0] ? Filename : L"<stdin>" );
	Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
	return 2;
}




unsigned int C2T_FileEncoding( LPTSTR Filename )
{
	if ( !Filename )
		return 0;


	unsigned int rv			= 0;
	_MMFile MMFile;


	if ( OpenMMFile( &MMFile, Filename, 0 ) == 0 ) {
		rv = MMFile.FileEncoding;
		CloseMMFile( &MMFile );
	}

	return rv;
}



int C2T_DumpBuffer( LPTSTR Buffer, int Chars, HANDLE OutHandle, BOOL Utf8 )
{
	char Utf8Buffer[4 * C2T_BUFFER_SIZE + 1]	= "";

	DWORD BytesToWrite			= 0;
	DWORD BytesWritten			= 0;
	int rv						= 0;


	if ( Utf8 ) {
		BytesToWrite = WideCharToMultiByte( CP_UTF8, 0, Buffer, Chars, Utf8Buffer, 4 * C2T_BUFFER_SIZE + 1, NULL, NULL );

		if ( BytesToWrite ) {
			if ( WriteFile( OutHandle, Utf8Buffer, BytesToWrite, &BytesWritten, NULL ) )
				rv = 0;
			else
				rv = GetLastError();
		}
		else
			rv = GetLastError();
	}
	else {
		if ( WriteFile( OutHandle, Buffer, Chars * sizeof( wchar_t ), &BytesWritten, NULL ) )
			rv = 0;
		else
			rv = GetLastError();
	}

	return rv;
}



DLLExports int WINAPI clip2text( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t OutBuffer[C2T_BUFFER_SIZE]		= L"";
	int BufferIndex							= 0;

	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";
	wchar_t AposWord[SHORT_BUF_LEN]			= L"";

	unsigned int Pause				= 0;
	BOOL Append						= FALSE;
	BOOL Overwrite					= FALSE;
	BOOL FixQuotes					= FALSE;
	BOOL Quiet						= FALSE;
	BOOL WriteBOM					= TRUE;
	BOOL Utf8						= FALSE;
	BOOL ToStdout					= TRUE;

	LARGE_INTEGER FileSize			= { 0 };
	DWORD CreateMode				= CREATE_NEW;
	unsigned int CurrentEncoding	= 0;

	HANDLE OutHandle				= NULL;
	HANDLE ClipDataHandle			= NULL;
	wchar_t *ClipData				= NULL;
	wchar_t *ch						= NULL;
	wchar_t thisch					= '\0';
	wchar_t lastch					= '\0';
	wchar_t nextch					= '\0';
	wchar_t ahead2					= '\0';
	unsigned __int64 CharsWritten	= 0;

	int rv							= 0;
	int erv							= 0;


	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( Clip2TextHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( TEXT2CLIP_VERSION ) );
			else if ( !_wcsicmp( arg[i], L"/A" ) )
				Append = TRUE, Overwrite = FALSE;
			else if ( !_wcsicmp( arg[i], L"/NB" ) || !_wcsicmp( arg[i], L"/N:B" ) )
				WriteBOM = FALSE;
			else if ( !_wcsicmp( arg[i], L"/O" ) )
				Overwrite = TRUE, Append = FALSE;
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					return 1;
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				FixQuotes = TRUE;
			else if ( !_wcsicmp( arg[i], L"/T" ) )
				Quiet = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
				int j = 2;

				if ( !_wcsnicmp( arg[i] + j, L"TF", 2 ) )
					j += 2;

				if ( arg[i][j] == '-' )
					j++;

				if ( !_wcsicmp( arg[i] + j, L"8" ) )
					Utf8 = TRUE;
				else if ( !_wcsicmp( arg[i] + j, L"16" ) )
					Utf8 = FALSE;
				else {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Unknown suboption: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( Filename[0] == '\0' ) {
				if ( QueryIsURL( arg[i] ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"URLs are not supported: \"%s\"\r\n", arg[i] );
					return 1;
				}
				else if ( !_wcsicmp( arg[i], ClipFilename ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"CLIP: is not a valid filename\r\n" );
					return 1;
				}
				else if ( !wcscmp( arg[i], StdinFilename ) ) {
					wcscpy_s( Filename, MAX_FILENAME_LEN, arg[i] );
					ToStdout = TRUE;
				}
				else {
					FH_Truename( arg[i], Filename );
					ToStdout = FALSE;
				}
			}
			else {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}


	if ( ToStdout && ConsoleHandle( STD_OUTPUT_HANDLE ) )
		WriteBOM = FALSE, Utf8 = FALSE;


	if ( FixQuotes )
		InitializeFixQuotesVars();

	if ( Pause )
		CL_SetPauseLines( Pause );


	if ( !IsClipboardFormatAvailable( CF_UNICODETEXT ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No text on the clipboard\r\n" );
		return 2;
	}


	if ( !ToStdout && !Overwrite && !Append ) {
		if ( QueryIsFile( Filename ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"File exists: \"%s\"\r\n", Filename );
			return 2;
		}
	}


	if ( Append )
		CreateMode = OPEN_ALWAYS;
	else if ( Overwrite )
		CreateMode = CREATE_ALWAYS;


	if ( !ToStdout && Append && QueryIsFile( Filename ) ) {
		CurrentEncoding = C2T_FileEncoding( Filename );

		if ( CurrentEncoding == FILE_UTF8 )
			Utf8 = TRUE;
		else if ( CurrentEncoding == FILE_UTF16LE )
			Utf8 = FALSE;
	}


	if ( !ToStdout ) {
		OutHandle = CreateFile( Filename, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CreateMode, FILE_ATTRIBUTE_NORMAL, NULL );
		
		if ( OutHandle == INVALID_HANDLE_VALUE ) {
			GetErrorText( GetLastError(), ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
			Qprintf( ERRHANDLE, L"   \"%s\"\r\n", Filename );
			return 2;
		}

		if ( Append ) {
			GetFileSizeEx( OutHandle, &FileSize );

			if ( FileSize.QuadPart ) {
				WriteBOM = FALSE;
				SetFilePointer( OutHandle, 0, NULL, FILE_END );
			}
		}
	}

	CL_HookBreak();


	if ( OpenClipboard( GetConsoleWindow() ) ) {
		ClipDataHandle = GetClipboardData( CF_UNICODETEXT );

		if ( ClipDataHandle ) {
			ClipData = (wchar_t*) GlobalLock( ClipDataHandle );

			if ( ClipData ) {
				ch = ClipData;

				if ( WriteBOM ) {
					OutBuffer[BufferIndex++] = BOM;
					OutBuffer[BufferIndex] = '\0';
					CharsWritten--;
				}

				while ( *ch ) {

					if ( _MMIsEOL( *ch ) && ToStdout ) {
						if ( *(ch + 1) == _MM2PartEOL( *ch ) )
							ch += 2;

						CharsWritten = CharsWritten + MM_CodePointsIn( OutBuffer ) + 2;

						QPuts( OutBuffer );
						if ( ( erv = CL_EndLine() ) == USER_ABORT )
							goto ErrorAbort;

						BufferIndex = 0;
						OutBuffer[BufferIndex] = '\0';
						lastch = '\n';
						continue;
					}

					if ( ( ( *ch & MM_SURROGATE_SIG_MASK ) == MM_SURROGATE_SIG_HIGH ) && ( BufferIndex >= ( C2T_BUFFER_SIZE - 2 ) ) ) {
						//	found a high surrogate, and there isn't enough room in the buffer for a surrogate pair:

						CharsWritten += MM_CodePointsIn( OutBuffer );

						if ( ToStdout )
							QPuts( OutBuffer );
						else {
							if ( erv = C2T_DumpBuffer( OutBuffer, BufferIndex, OutHandle, Utf8 ) )
								goto ErrorAbort;
						}

						BufferIndex = 0;
						OutBuffer[BufferIndex] = '\0';
					}


					thisch = *ch;

					if ( FixQuotes ) {
						nextch = *(ch + 1);
						if ( nextch )
							ahead2 = *(ch + 2);
						else
							ahead2 = '\0';

						if ( ReplaceDoubles && ( thisch == APOSTROPHE ) && ( nextch == APOSTROPHE ) ) {
							if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( ahead2 ) ) {
								thisch = QUOTE;
								ch++;
								nextch = ahead2;
							}
						}

						if ( ReplaceDoubles && ReplaceGraves && ( thisch == GRAVE ) && ( nextch == GRAVE ) ) {
							if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( ahead2 ) ) {
								thisch = LDQUOTE;
								ch++;
								nextch = ahead2;
							}
						}

						if ( ReplaceGraves && ( thisch == GRAVE ) )
							if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( ahead2 ) )
								thisch = LSQUOTE;

						if ( ( thisch == QUOTE ) || ( thisch == FWQUOTE ) ) {
							if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswspace( nextch ) || _MMIsEOL( nextch ) ) )
								thisch = DoublePrime;
							else if ( iswalpha( lastch ) && !iswalpha( nextch ) )
								thisch = CloseQuote;
							else if ( iswalpha( nextch ) && !iswalpha( lastch ) )
								thisch = OpenQuote;
							else if ( iswgraph( lastch ) && ( lastch != OpenSQuote ) )
								thisch = CloseQuote;
							else {
								thisch = OpenQuote;

								if ( iswspace( lastch ) )
									if ( _MMIsEOL( nextch ) )
										thisch = CloseQuote;
							}
						}
						else if ( ( thisch == APOSTROPHE ) || ( thisch == FWAPOSTROPHE ) ) {
							if ( iswgraph( lastch ) && ( lastch != OpenQuote ) ) {
								thisch = CloseSQuote;

								if ( iswalnum( lastch ) && iswalpha( nextch ) )
									thisch = Apostrophe;
							}
							else {
								thisch = OpenSQuote;

								if ( iswspace( lastch ) )
									if ( _MMIsEOL( nextch ) )
										thisch = CloseSQuote;
							}

							if ( !iswalnum( lastch ) && iswalpha( nextch ) ) {
								//		check for word beginning with apostrophe ....

								AposWord[0] = APOSTROPHE;
								int peek = 0;
								for ( int j = 0; j < 15; j++ ) {
									peek = *(ch + j + 1);
									AposWord[j + 1] = ( peek < SPACE ) ? SPACE : peek;
									if ( !peek )
										break;
								}

								AposWord[16] = '\0';
								if ( IsWordBeginningWithApostrophe( AposWord ) )
									thisch = Apostrophe;
							}
							else if ( IsVowel2( lastch ) && IsVowel2( nextch ) )
								thisch = Okina;
					
							if ( ( iswalnum( lastch ) == 0 ) && iswdigit( nextch ) )
								if ( iswdigit( ahead2 ) )
									if ( iswalnum( *(ch + 3) ) == 0 )
										thisch = Apostrophe;

							if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswdigit( nextch ) || IsFraction( nextch ) || iswspace( nextch ) || _MMIsEOL( nextch ) ) )
								thisch = Prime;
						}
						else if ( ( thisch == '-' ) || ( thisch == 0xff0d ) ) {
							if ( ( ( lastch != '-' ) && ( lastch != 0xff0d ) ) && ( nextch == thisch ) ) {
								if ( ( ahead2 != '-' ) && ( ahead2 != 0xff0d ) ) {
									thisch = EmDash;
									ch++;
								}
							}
						}
					}

					OutBuffer[BufferIndex++]	= thisch;
					OutBuffer[BufferIndex]		= '\0';

					if ( BufferIndex == ( C2T_BUFFER_SIZE - 1 ) ) {
						CharsWritten += MM_CodePointsIn( OutBuffer );

						if ( ToStdout )
							QPuts( OutBuffer );
						else {
							if ( erv = C2T_DumpBuffer( OutBuffer, BufferIndex, OutHandle, Utf8 ) )
								goto ErrorAbort;
						}

						BufferIndex = 0;
						OutBuffer[BufferIndex] = '\0';
					}

					lastch = *ch;
					ch++;
				}

				if ( BufferIndex ) {
					CharsWritten += MM_CodePointsIn( OutBuffer );

					if ( ToStdout )
						QPuts( OutBuffer );
					else {
						if ( erv = C2T_DumpBuffer( OutBuffer, BufferIndex, OutHandle, Utf8 ) )
							goto ErrorAbort;
					}

					BufferIndex = 0;
					OutBuffer[BufferIndex] = '\0';
				}

				GlobalUnlock( ClipDataHandle );
				ClipDataHandle = NULL;
			}
			else
				rv = GetLastError();
		}
		else
			rv = GetLastError();

		CloseClipboard();
	}
	else
		rv = GetLastError();



	if ( OutHandle )
		CloseHandle( OutHandle );


	if ( rv ) {
		GetErrorText( rv, ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
		rv = 2;
	}
	else if ( !Quiet ) {
		if ( ToStdout )
			if ( ( erv = CL_EndLine() == USER_ABORT ) )
				goto ErrorAbort;

		Printf( L"Wrote %s characters to ", Commafy( CharsWritten ) );

		if ( ToStdout )
			QPuts( L"<stdout>." );
		else
			Printf( L"\"%s\".", Filename );

		if ( ( erv = CL_EndLine() ) == USER_ABORT )
			goto ErrorAbort;
	}

	CL_ReleaseBreak();
	return rv;


ErrorAbort:

	if ( erv > 0 ) {
		GetErrorText( erv, ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
	}

	if ( OutHandle )
		CloseHandle( OutHandle );

	if ( ClipDataHandle && ClipData )
		GlobalUnlock( ClipDataHandle );

	CloseClipboard();
	CL_ReleaseBreak();

	if ( erv == USER_ABORT )
		return 3;
	else
		return 2;
}


#endif	//	TEXT2CLIP_VERSION


/*

	-			2021-10-15

	1.0.0.0		2023-07-27
				Adds a version number to this source file.  Help text tweaks.

	1.0.0.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
